@@ -9,7 +9,12 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
+#include <minizinc/ast.hh>
 #include <minizinc/flat_exp.hh>
+#include <minizinc/flatten_internal.hh>
+#include <minizinc/gc.hh>
+
+#include <vector>
 
 namespace MiniZinc {
 
@@ -20,7 +25,7 @@ std::vector<Expression*> get_conjuncts(Expression* start) {
   while (!conj_stack.empty()) {
     Expression* e = conj_stack.back();
     conj_stack.pop_back();
-    if (auto* bo = e->dynamicCast<BinOp>()) {
+    if (auto* bo = Expression::dynamicCast<BinOp>(e)) {
       if (bo->op() == BOT_AND) {
         conj_stack.push_back(bo->rhs());
         conj_stack.push_back(bo->lhs());
@@ -37,12 +42,13 @@ std::vector<Expression*> get_conjuncts(Expression* start) {
 void classify_conjunct(EnvI& env, Expression* e, IdMap<int>& eq_occurrences,
                        IdMap<std::pair<Expression*, Expression*>>& eq_branches,
                        std::vector<Expression*>& other_branches) {
-  if (auto* bo = e->dynamicCast<BinOp>()) {
-    if (bo->op() == BOT_EQ && bo->lhs()->type().dim() == 0) {
-      auto* ident = bo->lhs()->dynamicCast<Id>();
+  if (auto* bo = Expression::dynamicCast<BinOp>(e)) {
+    if (bo->op() == BOT_EQ && Expression::type(bo->lhs()).dim() == 0 &&
+        !Expression::type(bo->lhs()).structBT()) {
+      auto* ident = Expression::dynamicCast<Id>(bo->lhs());
       auto* other = bo->rhs();
       if (ident == nullptr || ident == env.constants.absent) {
-        ident = bo->rhs()->dynamicCast<Id>();
+        ident = Expression::dynamicCast<Id>(bo->rhs());
         other = bo->lhs();
       }
       if (ident != nullptr && ident != env.constants.absent &&
@@ -61,9 +67,68 @@ void classify_conjunct(EnvI& env, Expression* e, IdMap<int>& eq_occurrences,
   other_branches.push_back(e);
 }
 
+Expression* ite_struct_split(EnvI& env, Type ty, const std::vector<Expression*>& then_in,
+                             Expression* else_in, std::vector<KeepAlive>& results,
+                             std::vector<std::vector<KeepAlive>>& e_then,
+                             std::vector<KeepAlive>& e_else) {
+  StructType* st = env.getStructType(ty);
+  GCLock lock;
+  std::vector<Expression*> tupleResult(st->size());
+
+  std::vector<VarDecl*> then_decl(then_in.size());
+  for (int i = 0; i < then_in.size(); ++i) {
+    then_decl[i] =
+        new VarDecl(Location().introduce(),
+                    new TypeInst(Location().introduce(), Expression::type(then_in[i]), nullptr),
+                    env.genId(), then_in[i]);
+    then_decl[i]->ti()->setStructDomain(env, Expression::type(then_in[i]));
+  }
+  auto* else_decl =
+      new VarDecl(Location().introduce(),
+                  new TypeInst(Location().introduce(), Expression::type(else_in), nullptr),
+                  env.genId(), else_in);
+  else_decl->ti()->setStructDomain(env, Expression::type(else_in));
+
+  for (int i = 0; i < st->size(); ++i) {
+    Type field = (*st)[i];
+    if (field.structBT()) {
+      std::vector<Expression*> f_then_in(then_decl.size());
+      for (int j = 0; j < then_decl.size(); ++j) {
+        f_then_in[j] =
+            new FieldAccess(Location().introduce(), then_decl[i]->id(), IntLit::a(i + 1));
+        Expression::type(f_then_in[j], field);
+      }
+      Expression* f_else_in =
+          new FieldAccess(Location().introduce(), else_decl->id(), IntLit::a(i + 1));
+      tupleResult[i] = ite_struct_split(env, field, f_then_in, f_else_in, results, e_then, e_else);
+    } else {
+      VarDecl* fieldRes =
+          new_vardecl(env, Ctx(), new TypeInst(Location().introduce(), field, nullptr), nullptr,
+                      nullptr, nullptr);
+      tupleResult[i] = fieldRes->id();
+
+      results.emplace_back(fieldRes);
+      e_then.emplace_back();
+      e_then.back().reserve(then_decl.size());
+      for (auto& decl : then_decl) {
+        e_then.back().push_back(
+            new FieldAccess(Location().introduce(), decl->id(), IntLit::a(i + 1)));
+        Expression::type(e_then.back().back()(), field);
+      }
+      e_else.emplace_back(
+          new FieldAccess(Location().introduce(), else_decl->id(), IntLit::a(i + 1)));
+      Expression::type(e_else.back()(), field);
+    }
+    assert(results.size() == e_then.size() && results.size() == e_else.size());
+  }
+  ArrayLit* tuple = ArrayLit::constructTuple(Location().introduce(), tupleResult);
+  tuple->type(ty);
+  return tuple;
+}
+
 EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b) {
   CallStackItem _csi(env, e);
-  ITE* ite = e->cast<ITE>();
+  ITE* ite = Expression::cast<ITE>(e);
 
   // The conditions of each branch of the if-then-else
   std::vector<KeepAlive> conditions;
@@ -86,7 +151,8 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
   bool allDefined = true;
 
   // The result variables of each generated conditional
-  std::vector<VarDecl*> results;
+  std::vector<KeepAlive> results;
+  KeepAlive tupleResult;
   // The then-expressions of each generated conditional
   std::vector<std::vector<KeepAlive>> e_then;
   // The else-expressions of each generated conditional
@@ -116,15 +182,15 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
     for (auto& eq : eq_occurrences) {
       if (eq.second >= ite->size()) {
         // Any identifier that occurs in all or all but one branch gets its own conditional
-        results.push_back(eq.first->decl());
+        results.emplace_back(eq.first->decl());
         e_then.emplace_back();
         for (int i = 0; i < ite->size(); i++) {
           auto it = eq_branches[i].find(eq.first);
           if (it == eq_branches[i].end()) {
             // not found, simply push x=x
-            e_then.back().push_back(eq.first);
+            e_then.back().emplace_back(eq.first);
           } else {
-            e_then.back().push_back(it->second.first);
+            e_then.back().emplace_back(it->second.first);
           }
         }
         {
@@ -149,23 +215,23 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
       }
     }
     if (!noOtherBranches) {
-      results.push_back(r);
+      results.emplace_back(r);
       e_then.emplace_back();
       for (int i = 0; i < ite->size(); i++) {
         if (eq_branches[i].empty()) {
-          e_then.back().push_back(ite->thenExpr(i));
+          e_then.back().emplace_back(ite->thenExpr(i));
         } else if (other_branches[i].empty()) {
-          e_then.back().push_back(env.constants.literalTrue);
+          e_then.back().emplace_back(env.constants.literalTrue);
         } else if (other_branches[i].size() == 1) {
-          e_then.back().push_back(other_branches[i][0]);
+          e_then.back().emplace_back(other_branches[i][0]);
         } else {
           GCLock lock;
           auto* al = new ArrayLit(Location().introduce(), other_branches[i]);
           al->type(Type::varbool(1));
-          Call* forall = new Call(Location().introduce(), env.constants.ids.forall, {al});
+          Call* forall = Call::a(Location().introduce(), env.constants.ids.forall, {al});
           forall->decl(env.model->matchFn(env, forall, false));
-          forall->type(forall->decl()->rtype(env, {al}, false));
-          e_then.back().push_back(forall);
+          forall->type(forall->decl()->rtype(env, {al}, nullptr, false));
+          e_then.back().emplace_back(forall);
         }
       }
       {
@@ -179,19 +245,27 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
           GCLock lock;
           auto* al = new ArrayLit(Location().introduce(), other_branches[ite->size()]);
           al->type(Type::varbool(1));
-          Call* forall = new Call(Location().introduce(), env.constants.ids.forall, {al});
+          Call* forall = Call::a(Location().introduce(), env.constants.ids.forall, {al});
           forall->decl(env.model->matchFn(env, forall, false));
-          forall->type(forall->decl()->rtype(env, {al}, false));
+          forall->type(forall->decl()->rtype(env, {al}, nullptr, false));
           e_else.emplace_back(forall);
         }
       }
     }
+  } else if (ite->type().istuple() || ite->type().isrecord()) {
+    noOtherBranches = false;
+    std::vector<Expression*> then_in(ite->size());
+    for (size_t i = 0; i < ite->size(); ++i) {
+      then_in[i] = ite->thenExpr(i);
+    }
+    tupleResult =
+        ite_struct_split(env, ite->type(), then_in, ite->elseExpr(), results, e_then, e_else);
   } else {
     noOtherBranches = false;
-    results.push_back(r);
+    results.emplace_back(r);
     e_then.emplace_back();
     for (int i = 0; i < ite->size(); i++) {
-      e_then.back().push_back(ite->thenExpr(i));
+      e_then.back().emplace_back(ite->thenExpr(i));
     }
     e_else.emplace_back(ite->elseExpr());
   }
@@ -220,8 +294,8 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
   for (int i = 0; i < ite->size() && !foundTrueBranch; i++) {
     bool cond = true;
     EE e_if;
-    if (ite->ifExpr(i)->isa<Call>() &&
-        ite->ifExpr(i)->cast<Call>()->id() == env.constants.ids.mzn_in_root_context) {
+    if (Expression::isa<Call>(ite->ifExpr(i)) &&
+        Expression::cast<Call>(ite->ifExpr(i))->id() == env.constants.ids.mzn_in_root_context) {
       e_if = EE(env.constants.boollit(ctx.b == C_ROOT), env.constants.literalTrue);
     } else {
       Ctx cmix_not_negated;
@@ -229,7 +303,7 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
       cmix_not_negated.i = C_MIX;
       e_if = flat_exp(env, cmix_not_negated, ite->ifExpr(i), nullptr, env.constants.varTrue);
     }
-    if (e_if.r()->type() == Type::parbool()) {
+    if (Expression::type(e_if.r()) == Type::parbool()) {
       {
         GCLock lock;
         cond = eval_bool(env, e_if.r());
@@ -250,7 +324,7 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
           defined[j].push_back(ethen.b);
           allDefined = allDefined && (ethen.b() == env.constants.literalTrue);
           branches[j].push_back(ethen.r);
-          if (ethen.r()->type().isvar()) {
+          if (Expression::type(ethen.r()).isvar()) {
             allBranchesPar[j] = false;
           }
         }
@@ -259,8 +333,8 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
         GCLock lock;
         conditions.emplace_back(env.constants.literalFalse);
         for (unsigned int j = 0; j < results.size(); j++) {
-          defined[j].push_back(env.constants.literalTrue);
-          branches[j].push_back(create_dummy_value(env, e_then[j][i]()->type()));
+          defined[j].emplace_back(env.constants.literalTrue);
+          branches[j].emplace_back(create_dummy_value(env, Expression::type(e_then[j][i]())));
         }
       }
     } else {
@@ -276,7 +350,7 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
         defined[j].push_back(ethen.b);
         allDefined = allDefined && (ethen.b() == env.constants.literalTrue);
         branches[j].push_back(ethen.r);
-        if (ethen.r()->type().isvar()) {
+        if (Expression::type(ethen.r()).isvar()) {
           allBranchesPar[j] = false;
         }
       }
@@ -285,21 +359,21 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
 
     if (cond) {
       for (unsigned int j = 0; j < results.size(); j++) {
-        if (r_bounds_valid_int[j] && e_then[j][i]()->type().isint()) {
+        if (r_bounds_valid_int[j] && Expression::type(e_then[j][i]()).isint()) {
           GCLock lock;
           IntBounds ib_then = compute_int_bounds(env, branches[j][i]());
           if (ib_then.valid) {
             r_bounds_int[j].push_back(ib_then);
           }
           r_bounds_valid_int[j] = r_bounds_valid_int[j] && ib_then.valid;
-        } else if (r_bounds_valid_set[j] && e_then[j][i]()->type().isIntSet()) {
+        } else if (r_bounds_valid_set[j] && Expression::type(e_then[j][i]()).isIntSet()) {
           GCLock lock;
           IntSetVal* isv = compute_intset_bounds(env, branches[j][i]());
           if (isv != nullptr) {
             r_bounds_set[j].push_back(isv);
           }
           r_bounds_valid_set[j] = r_bounds_valid_set[j] && (isv != nullptr);
-        } else if (r_bounds_valid_float[j] && e_then[j][i]()->type().isfloat()) {
+        } else if (r_bounds_valid_float[j] && Expression::type(e_then[j][i]()).isfloat()) {
           GCLock lock;
           FloatBounds fb_then = compute_float_bounds(env, branches[j][i]());
           if (fb_then.valid) {
@@ -318,7 +392,7 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
   }
 
   for (auto& result : results) {
-    if (result == nullptr) {
+    if (result() == nullptr) {
       // need to introduce new result variable
       GCLock lock;
       auto* ti = new TypeInst(Location().introduce(), ite->type(), nullptr);
@@ -337,25 +411,25 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
       defined[j].push_back(eelse.b);
       allDefined = allDefined && (eelse.b() == env.constants.literalTrue);
       branches[j].push_back(eelse.r);
-      if (eelse.r()->type().isvar()) {
+      if (Expression::type(eelse.r()).isvar()) {
         allBranchesPar[j] = false;
       }
 
-      if (r_bounds_valid_int[j] && e_else[j]()->type().isint()) {
+      if (r_bounds_valid_int[j] && Expression::type(e_else[j]()).isint()) {
         GCLock lock;
         IntBounds ib_else = compute_int_bounds(env, eelse.r());
         if (ib_else.valid) {
           r_bounds_int[j].push_back(ib_else);
         }
         r_bounds_valid_int[j] = r_bounds_valid_int[j] && ib_else.valid;
-      } else if (r_bounds_valid_set[j] && e_else[j]()->type().isIntSet()) {
+      } else if (r_bounds_valid_set[j] && Expression::type(e_else[j]()).isIntSet()) {
         GCLock lock;
         IntSetVal* isv = compute_intset_bounds(env, eelse.r());
         if (isv != nullptr) {
           r_bounds_set[j].push_back(isv);
         }
         r_bounds_valid_set[j] = r_bounds_valid_set[j] && (isv != nullptr);
-      } else if (r_bounds_valid_float[j] && e_else[j]()->type().isfloat()) {
+      } else if (r_bounds_valid_float[j] && Expression::type(e_else[j]()).isfloat()) {
         GCLock lock;
         FloatBounds fb_else = compute_float_bounds(env, eelse.r());
         if (fb_else.valid) {
@@ -367,9 +441,8 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
   }
 
   // update domain of result variable with bounds from all branches
-
   for (unsigned int j = 0; j < results.size(); j++) {
-    VarDecl* nr = results[j];
+    auto* nr = Expression::cast<VarDecl>(results[j]());
     GCLock lock;
     if (r_bounds_valid_int[j] && ite->type().isint()) {
       IntVal lb = IntVal::infinity();
@@ -440,69 +513,89 @@ EE flatten_ite(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, VarDecl* b)
   }
 
   // Create ite predicate calls
-  GCLock lock;
-  auto* al_cond = new ArrayLit(Location().introduce(), conditions);
-  al_cond->type(Type::varbool(1));
+  KeepAlive al_cond;
+  {
+    GCLock lock;
+    al_cond = new ArrayLit(Location().introduce(), conditions);
+    Expression::type(al_cond(), Type::varbool(1));
+  }
   for (unsigned int j = 0; j < results.size(); j++) {
-    auto* al_branches = new ArrayLit(Location().introduce(), branches[j]);
-    Type branches_t = results[j]->type();
-    branches_t.dim(1);
-    branches_t.ti(allBranchesPar[j] ? Type::TI_PAR : Type::TI_VAR);
-    al_branches->type(branches_t);
-    Call* ite_pred = new Call(ite->loc().introduce(), ASTString("if_then_else"),
-                              {al_cond, al_branches, results[j]->id()});
-    ite_pred->decl(env.model->matchFn(env, ite_pred, false));
-    ite_pred->type(Type::varbool());
-    make_defined_var(env, results[j], ite_pred);
-    (void)flat_exp(env, Ctx(), ite_pred, env.constants.varTrue, env.constants.varTrue);
+    KeepAlive ite_pred;
+    {
+      GCLock lock;
+      auto* al_branches = new ArrayLit(Location().introduce(), branches[j]);
+      Type branches_t = Type::arrType(env, Type::bot(1), Expression::type(results[j]()));
+      if (!allBranchesPar[j]) {
+        branches_t.mkVar(env);
+      }
+      branches_t.ti(allBranchesPar[j] ? Type::TI_PAR : Type::TI_VAR);
+      al_branches->type(branches_t);
+      ite_pred = Call::a(Expression::loc(ite).introduce(), ASTString("if_then_else"),
+                         {al_cond(), al_branches, Expression::cast<VarDecl>(results[j]())->id()});
+      Expression::cast<Call>(ite_pred())
+          ->decl(env.model->matchFn(env, Expression::cast<Call>(ite_pred()), false));
+      Expression::type(Expression::cast<Call>(ite_pred()), Type::varbool());
+      make_defined_var(env, Expression::cast<VarDecl>(results[j]()),
+                       Expression::cast<Call>(ite_pred()));
+    }
+    (void)flat_exp(env, Ctx(), ite_pred(), env.constants.varTrue, env.constants.varTrue);
   }
   EE ret;
   if (noOtherBranches) {
     ret.r = env.constants.varTrue->id();
+  } else if (ite->type().istuple() || ite->type().isrecord()) {
+    ret.r = bind(env, ctx, r, tupleResult());
   } else {
-    ret.r = results.back()->id();
+    ret.r = Expression::cast<VarDecl>(results.back()())->id();
   }
   if (allDefined) {
     bind(env, Ctx(), b, env.constants.literalTrue);
     ret.b = env.constants.literalTrue;
   } else {
     // Otherwise, constraint linking conditions, b and the definedness variables
-    if (b == nullptr) {
-      CallStackItem _csi(env, new StringLit(Location().introduce(), "b"));
-      b = new_vardecl(env, Ctx(), new TypeInst(Location().introduce(), Type::varbool()), nullptr,
-                      nullptr, nullptr);
-    }
-    ret.b = b->id();
+    KeepAlive ite_defined_pred;
+    {
+      GCLock lock;
 
-    std::vector<Expression*> defined_conjunctions(conditions.size());
-    for (unsigned int i = 0; i < conditions.size(); i++) {
-      std::vector<Expression*> def_i;
-      for (auto& j : defined) {
-        assert(j.size() > i);
-        if (j[i]() != env.constants.literalTrue) {
-          def_i.push_back(j[i]());
+      if (b == nullptr) {
+        CallStackItem _csi(env, new StringLit(Location().introduce(), "b"));
+        b = new_vardecl(env, Ctx(), new TypeInst(Location().introduce(), Type::varbool()), nullptr,
+                        nullptr, nullptr);
+      }
+      ret.b = b->id();
+
+      std::vector<Expression*> defined_conjunctions(conditions.size());
+      for (unsigned int i = 0; i < conditions.size(); i++) {
+        std::vector<Expression*> def_i;
+        for (auto& j : defined) {
+          assert(j.size() > i);
+          if (j[i]() != env.constants.literalTrue) {
+            def_i.push_back(j[i]());
+          }
+        }
+        if (def_i.empty()) {
+          defined_conjunctions[i] = env.constants.literalTrue;
+        } else if (def_i.size() == 1) {
+          defined_conjunctions[i] = def_i[0];
+        } else {
+          auto* al = new ArrayLit(Location().introduce(), def_i);
+          al->type(Type::varbool(1));
+          Call* forall = Call::a(Location().introduce(), env.constants.ids.forall, {al});
+          forall->decl(env.model->matchFn(env, forall, false));
+          forall->type(forall->decl()->rtype(env, {al}, nullptr, false));
+          defined_conjunctions[i] = forall;
         }
       }
-      if (def_i.empty()) {
-        defined_conjunctions[i] = env.constants.literalTrue;
-      } else if (def_i.size() == 1) {
-        defined_conjunctions[i] = def_i[0];
-      } else {
-        auto* al = new ArrayLit(Location().introduce(), def_i);
-        al->type(Type::varbool(1));
-        Call* forall = new Call(Location().introduce(), env.constants.ids.forall, {al});
-        forall->decl(env.model->matchFn(env, forall, false));
-        forall->type(forall->decl()->rtype(env, {al}, false));
-        defined_conjunctions[i] = forall;
-      }
+      auto* al_defined = new ArrayLit(Location().introduce(), defined_conjunctions);
+      al_defined->type(Type::varbool(1));
+      ite_defined_pred =
+          Call::a(Expression::loc(ite).introduce(), ASTString("if_then_else_partiality"),
+                  {al_cond(), al_defined, b->id()});
+      Expression::cast<Call>(ite_defined_pred())
+          ->decl(env.model->matchFn(env, Expression::cast<Call>(ite_defined_pred()), false));
+      Expression::type(ite_defined_pred(), Type::varbool());
     }
-    auto* al_defined = new ArrayLit(Location().introduce(), defined_conjunctions);
-    al_defined->type(Type::varbool(1));
-    Call* ite_defined_pred = new Call(ite->loc().introduce(), ASTString("if_then_else_partiality"),
-                                      {al_cond, al_defined, b->id()});
-    ite_defined_pred->decl(env.model->matchFn(env, ite_defined_pred, false));
-    ite_defined_pred->type(Type::varbool());
-    (void)flat_exp(env, Ctx(), ite_defined_pred, env.constants.varTrue, env.constants.varTrue);
+    (void)flat_exp(env, Ctx(), ite_defined_pred(), env.constants.varTrue, env.constants.varTrue);
   }
 
   return ret;
