@@ -360,6 +360,8 @@ void remove_deleted_items(EnvI& envi, std::vector<VarDecl*>& deletedVarDecls) {
     if (envi.varOccurrences.occurrences(cur) == 0) {
       auto cur_idx = envi.varOccurrences.idx.find(cur->id());
       if (cur_idx.first && !m[*cur_idx.second]->removed()) {
+        auto* vdi = m[*cur_idx.second]->cast<VarDeclI>();
+        cur = vdi->e();
         if (is_output(cur)) {
           // We have to change the output model if we remove this variable
           Expression* val = nullptr;
@@ -399,16 +401,14 @@ void remove_deleted_items(EnvI& envi, std::vector<VarDecl*>& deletedVarDecls) {
             VarDecl* vd_out =
                 (*envi.output)[envi.outputFlatVarOccurrences.find(cur)]->cast<VarDeclI>()->e();
             vd_out->e(val);
-            CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls,
-                            m[*cur_idx.second]->cast<VarDeclI>());
+            CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls, vdi);
             top_down(cd, cur->e());
-            (*envi.flat())[*cur_idx.second]->remove();
+            vdi->remove();
           }
         } else {
-          CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls,
-                          m[*cur_idx.second]->cast<VarDeclI>());
+          CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls, vdi);
           top_down(cd, cur->e());
-          (*envi.flat())[*cur_idx.second]->remove();
+          vdi->remove();
         }
       }
     }
@@ -1532,7 +1532,22 @@ bool simplify_constraint(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVarD
           } else {
             CollectDecls cd(env, env.varOccurrences, deletedVarDecls, ii);
             top_down(cd, c);
-            ii->remove();
+            if (auto* vdi = ii->dynamicCast<VarDeclI>()) {
+              auto* v = IntLit::a(b2i_val ? 1 : 0);
+              if (env.varOccurrences.occurrences(vdi->e()) == 0) {
+                if (is_output(vdi->e())) {
+                  VarDecl* vd_out = (*env.output)[env.outputFlatVarOccurrences.find(vdi->e())]
+                                        ->cast<VarDeclI>()
+                                        ->e();
+                  vd_out->e(v);
+                }
+                vdi->remove();
+              } else {
+                vdi->e()->e(v);
+              }
+            } else {
+              ii->remove();
+            }
           }
         } else {
           Id* ident = Expression::cast<Id>(c->arg(0));
@@ -2093,7 +2108,9 @@ void substitute_fixed_vars(Env& env) {
     Model& m = *envi.flat();
     std::vector<VarDecl*> deletedVarDecls;
     for (auto* item : m) {
-      substitute_fixed_vars(envi, item, deletedVarDecls);
+      if (!item->removed()) {
+        substitute_fixed_vars(envi, item, deletedVarDecls);
+      }
     }
     remove_deleted_items(envi, deletedVarDecls);
   } catch (ModelInconsistent&) { /* NOLINT(bugprone-empty-catch) */
