@@ -89,13 +89,14 @@ vector<string> MIPGurobiWrapper::getStdFlags() { return {"-i", "-p", "-s", "-v"}
 
 vector<string> gurobi_dlls() {
   const vector<string> versions = {
-      "954", "953", "952", "951",  // Potential future versions which should load correctly
-      "950", "912", "911", "910", "903", "902", "901", "900", "811", "810", "801",
-      "800", "752", "751", "750", "702", "701", "700", "652", "651", "650"};
+      "1103", "1102", "1101",  // Potential future versions which should load correctly
+      "1100", "1003", "1002", "1001", "1000", "952", "951", "950", "912", "911",
+      "910",  "903",  "902",  "901",  "900",  "811", "810", "801", "800", "752",
+      "751",  "750",  "702",  "701",  "700",  "652", "651", "650"};
   vector<string> dlls;
   string lastMajorVersion;
   for (const auto& version : versions) {
-    string majorVersion = version.substr(0, 2);
+    string majorVersion = version.substr(0, version.length() - 1);
     if (majorVersion != lastMajorVersion) {
       dlls.push_back("gurobi" + majorVersion);
       lastMajorVersion = majorVersion;
@@ -104,6 +105,8 @@ vector<string> gurobi_dlls() {
     dlls.push_back("C:\\gurobi" + version + "\\win64\\bin\\gurobi" + majorVersion + ".dll");
 #elif __APPLE__
     dlls.push_back("/Library/gurobi" + version + "/mac64/lib/libgurobi" + majorVersion + ".dylib");
+    dlls.push_back("/Library/gurobi" + version + "/macos_universal2/lib/libgurobi" + majorVersion +
+                   ".dylib");
 #else
     dlls.push_back("/opt/gurobi" + version + "/linux64/lib/libgurobi" + majorVersion + ".so");
 #endif
@@ -193,7 +196,8 @@ bool MIPGurobiWrapper::Options::processOption(int& i, std::vector<std::string>&
   std::string buf;
   if (cop.get("-i")) {
     flagIntermediate = true;
-  } else if (string(argv[i]) == "-f") {              // NOLINT: Allow repeated empty if
+  } else if (string(argv[i]) == "-f" ||
+             string(argv[i]) == "--free-search") {   // NOLINT: Allow repeated empty if
   } else if (string(argv[i]) == "--fixed-search") {  // NOLINT: Allow repeated empty if
     nFreeSearch = MIPGurobiWrapper::SearchType::FIXED_SEARCH;
   } else if (string(argv[i]) == "--uniform-search") {  // NOLINT: Allow repeated empty if
